/*
 * File:	izp_project_00.c
 * Date:	11-15-2004
 * Auth:	-tHE SWINe-
 * Proj:	IZP no.1 - numeric system conversions
 * Desc:	Conversion of desired number from/to numeric system
 *
 * tab size 4
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdbool.h>
#include <ctype.h>
#include <string.h>
#include <assert.h>
#include <limits.h>

/*enum {
	false = 0,
	true = 1
};*/

// min and max numeric base
enum {
	Base_Min = 2,
	Base_Max = 36
};

/*
 *		void PrintHelp() - print helpscreen
 */
void PrintHelp()
{
	printf("                              === Help ===\n\n"
		   "izp_project_00 - Numeric systems\n"
		   "Auth: -tHE SWINe-\n\n"
		   "    Used for converting numbers from desired numeric system to another.\n"
		   "Number(s) to convert are read from standard input (stdin). Numbers > 9\n"
		   "are represented by great letters (A-Z; like hex)\n\n"
		   "Ussage: izp_project_00.c -h\n"
		   "        izp_project_00.c b1 b2\n\n"
		   "Parameter description:\n"
		   "  -h    Display this help screen\n"
		   "  b1    Input numeric base (2-36)\n"
		   "  b2    Output numeric base (2-36)\n\n");
}

/*
 *		void *p_testAlloc(void *p_buffer) - print error message if <p_buffer> == NULL.
 */
__inline void *p_testAlloc(void *p_buffer)
{
	if(p_buffer == NULL) {
		printf("Error - not enough memory");
		getchar();
		exit(-1);
	}

	return p_buffer;
}

// size of blocks read_line is incremented
enum {
	RL_Block_Increment = 16
};

/*
 *		unsigned int read_line(char **p_s_line, unsigned int *n_buffer_len, FILE *p_stream)
 *			- read line from stream <p_stream>
 *			- alloc memory for string (containing '\n'), fill <p_s_line> whith pointer to data
 *			  and <n_buffer_len> with buffer lenght (not string length !!!)
 *			- return true if success, otherwise false
 *			- accept <*p_s_line> != 0 && <*n_buffer_len> == allocated_memory_size
 */
unsigned int read_line(char **p_s_line, unsigned int *n_buffer_len, FILE *p_stream)
{
	int n_buffer_alloc;
	int c;
	int i;

	assert(p_s_line != NULL);
	assert(n_buffer_len != NULL);
	assert(p_stream != NULL);
	// param check

	if(*p_s_line == NULL) {
		n_buffer_alloc = RL_Block_Increment;
	    if((*p_s_line = malloc(n_buffer_alloc * sizeof(char))) == NULL)
			return false;
	} else
		n_buffer_alloc = *n_buffer_len;
	// alloc memory

	i = 0;
	while((c = getc(p_stream)) != EOF && c != '\n') {
		(*p_s_line)[i ++] = c;

		if(i >= n_buffer_alloc) {
			n_buffer_alloc += RL_Block_Increment;
			if((*p_s_line = realloc(*p_s_line, n_buffer_alloc * sizeof(char))) == NULL)
				return false;
		}
		// grow buffer if necessary (always have buffer size +1 for terminating 0)
	}
	// read string

	(*p_s_line)[i ++] = 0;
	// finish string

	*n_buffer_len = n_buffer_alloc;
	// memory size !!! not string length !!

	return true;
}

unsigned long n_Number(char *p_s_number, int n_base);
void Print_Number(unsigned long n_number, int n_base);

/*
 *		int main(int argc, char *argv[]) - main
 */
int main(int argc, char *argv[])
{
	int n_base_from, n_base_to;
	unsigned long n_number;
	unsigned int n_buffer_size;
	char *p_s_line;
	unsigned int i;

	if(argc == 2 && !stricmp(argv[1], "-h")) {
		PrintHelp();
		getchar();
		return 0;
	}
	// helpscreen request ...

	if(argc != 3) {
		printf("Error - not enough parameters");
		//
		PrintHelp();
		getchar();
		return -1;
	}
	//
	n_base_from = atol(argv[1]);
	n_base_to = atol(argv[2]);
	//
	if(n_base_from < 2 || n_base_from > 36 || n_base_to < 2 || n_base_to > 36) {
		printf("Error - min and max base must be within <2, 36>\n");
		//
		PrintHelp();
		getchar();
		return -1;
	}
	// check params ...

	p_s_line = 0;
	n_buffer_size = 0;
	while(1) {
		if(!read_line(&p_s_line, &n_buffer_size, stdin)) {
			printf("Error - not enough memory\n");
			getchar();
			return -1;
		}
		// read number from stdin as single line ...

		if(!strlen(p_s_line))
			break;
		// enter nothing to exit

		for(i = 0; i < strlen(p_s_line); i ++) {
			if(isspace(p_s_line[i]))
				p_s_line[i] = 0;
		}
		// erase whitespaces past number ...

		n_number = n_Number(p_s_line, n_base_from);
		// read number from string

		//printf("\n%u\n\n", n_number); // debug

		Print_Number(n_number, n_base_to);
		// print converted number

		printf("\n----------------\n");
	}

	free((void*)p_s_line);
	// don't forget !

	return 0;
}

/*
 *		void Error(char *p_s_error) - print error message <p_s_error> and exit with -1
 */
void Error(char *p_s_error)
{
	assert(p_s_error);

	printf("Error - %s\n", p_s_error);
	getchar();
	exit(-1);
}

/*
 *		int n_Char_Value(char c) - return value of number, represented by <c>.
 *								   must be number / lower alpha
 */
__inline int n_Char_Value(char c)
{
	return (c >= '0' && c <= '9')? c  - '0' : c - 'a' + 10;
}

/*
 *		unsigned long n_Number(char *p_s_number, int n_base)
 *			- calc value of number, represented by number <p_s_number> of numeric base <n_base>
 */
unsigned long n_Number(char *p_s_number, int n_base)
{
	unsigned long n_result, n_prev, n_base_pwr;
	int i;

	assert(n_base >= 2 && n_base <= 36);

	n_result = 0;

	for(i = 0; i < (signed)strlen(p_s_number); i ++) {
		if(isalpha(p_s_number[i])) {
			p_s_number[i] |= 32;
			// tolower

			if(p_s_number[i] > n_base - 11 + 'a')
				Error("Illegal alpha characters in number");
		} else if(isdigit(p_s_number[i])) {
			if(p_s_number[i] > n_base + '0')
				Error("Illegal digit in number");
		} else
			Error("Illegal characters in number");
	}
	// check number for illegal characters & make alpha characters lower

	for(i = (signed)strlen(p_s_number) - 1, n_result = 0, n_prev = 0, n_base_pwr = 1;
		i >= 0; i --, n_base_pwr *= n_base /* should never overflow */, n_prev = n_result) {

		n_result += n_Char_Value(p_s_number[i]) * n_base_pwr;

		if(n_prev > n_result) {
			Error("Number too big for poor 32-bit baby ...");
			// overflow !!!
		}
		// number size check ...

		if(i && (n_base_pwr > (ULONG_MAX / n_base) ||
		   (n_base_pwr == (ULONG_MAX / n_base)) && ULONG_MAX % n_base)) {
			Error("Number too big for poor 32-bit baby ...");
			// overflow !!!
		}
		// base_power size check ...
	}
	// main conversion loop

	return n_result;
}

/*
 *		__inline char n_Character(unsigned long n_number) - convert number to character
 */
__inline char n_Character(unsigned long n_number)
{
	if(n_number >= 36)
		Error("internal error - converting digit greater than 35");
	return (n_number < 10)? '0' + (char)n_number : 'A' + (char)n_number - 10;
}

enum {
	PN_Block_Increment = 16
};

/*
 *		void Print_Number(unsigned long n_number, int n_base)
 *			- print number <n_number>; result base = <n_base>
 */
void Print_Number(unsigned long n_number, int n_base)
{
	int n_buffer_size;
	char *p_buffer;
	int i;

	if(!n_number) {
		printf("0");
		return;
	}
	// zero is explicit ...

	p_testAlloc(p_buffer = (char*)malloc((n_buffer_size = PN_Block_Increment) * sizeof(char)));
	// alloc memory

	i = 0;
	while(n_number) {
		p_buffer[i ++] = n_Character(n_number % n_base);
		//
		n_number /= n_base;
		// number conversion

		if(i >= n_buffer_size) {
			n_buffer_size += PN_Block_Increment;
			if((p_buffer = realloc(p_buffer, n_buffer_size * sizeof(char))) == NULL)
				Error("not enough memory");
		}
		// grow buffer if necessary (always have buffer size +1 for terminating 0)
	}
	// read string

	p_buffer[i] = 0;
	// finish string

	for(i --; i >= 0; i --)
		putchar(p_buffer[i]);
	// print in reversed order

	free((void*)p_buffer);
	// don't forget !
}
